package arimaa3;

// You must rename this file to: TournamentSimulation.java before compiling

//import ai_util.*;

import java.io.*;
import java.util.*;

public class TournamentSimulation {


  public TournamentSimulation(String state_file_name) throws
    FileNotFoundException {
    read_file(state_file_name);
  }


  // Reads the tournament state file
  private class TournamentStateFileReader {

    protected BufferedReader br;
    protected String file_name;

    public TournamentStateFileReader(String file_name) throws
      FileNotFoundException {
      // Open the file
      this.file_name = file_name;
      FileReader fr = new FileReader(file_name);
      br = new BufferedReader(fr);
    }

    // returns true iff another position is available
    public boolean hasNext() {
      try {
        br.mark(2000);
        String result = readLine();
        br.reset();
        if (result != null) {
          return true;
        }
      } catch (IOException ex) {
      }
      return false;
    }

    // returns the next item, else throws exception
    public Object next() {
      try {
        String result = readLine();
        if (result != null) {
          return result;
        }
      } catch (IOException ex) {
      }
      throw new NoSuchElementException();
    }

    public void remove() {
      throw new UnsupportedOperationException();
    }

    // resets file reader to start of file
    public void reset() {
      try {
        br.close();
        // Open the file
        FileReader fr = new FileReader(file_name);
        br = new BufferedReader(fr);
      } catch (IOException ex) {
      }
    }


    // Reads a line from the file
    // returns null if no more data available
    protected String readLine() throws IOException {
      String text = null;

      while (true) {

        text = br.readLine();

        // Check for good data
        if (text == null) {
          break;
        }

        // Check if its a comment or a blank line
        text = text.trim();

        // Check for a blank line or a comment
        if (!text.startsWith("*") && !text.equals("") && !text.startsWith("#")) {
          break;
        }

      }

      return text;
    }

  }


  // Helper Function
  // Converts a text name to a player object
  private Player getPlayer(String player_text) {
    for (int i = 0; i < player_list.size(); i++) {
      Player player = (Player) player_list.get(i);
      if (player_text.equals(player.name)) {
        return player;
      }
    }

    if (player_text.equals("BYE")) {
      return bye;
    }

    assert (false);
    return null;
  }

  // Contains the results of a match
  class MatchResult {

    public MatchResult(String player_a, String player_b, String winner) {
      String t_loser = (player_a.equals(winner)) ? player_b : player_a;

      this.white = getPlayer(player_a);
      this.black = getPlayer(player_b);
      this.loser = getPlayer(t_loser);
      this.winner = getPlayer(winner);
    }

    public Player white;
    public Player black;
    public Player winner;
    public Player loser;

    public String toString() {
      return winner + " beat " + loser;
    }
  }


  Player bye = new Player("BYE", -1);

  // Contains the name and rating info of a player
  class Player implements Comparable {
    public Player(String name, int rating) {
      this.name = new String(name);
      this.rating = rating;
    }

    public String toString() {
      return name + " " + rating;
    }

    public int compareTo(Object o) {
      Player that = (Player) o;
      return (that.rating - this.rating);
    }

    public String name;
    public int rating;
    public int wins;
  }


  ArrayList player_list = new ArrayList();
  ArrayList match_list = new ArrayList();


  // processes the tournament state file info
  private void read_file(String ts_file) throws FileNotFoundException {
    TournamentStateFileReader reader = new TournamentStateFileReader(ts_file);

    while (reader.hasNext()) {
      String text = (String) reader.next();
//      LogFile.write(text);

      String[] words = text.split(" ");

      // Get the player list with measured ratings
      if (text.startsWith("player")) {
        Player temp_p = new Player(words[1], Integer.parseInt(words[2]));
        player_list.add(temp_p);
      }

      // get the match results
      if (text.startsWith("pick")) {
        MatchResult temp_r = new MatchResult(words[1], words[2], words[4]);
        match_list.add(temp_r);
      }

      // "done" must be written on a separate invocation
      if (text.startsWith("result") && text.endsWith("1")) {
        System.out.println("done");
        return;
      }
    }

  }


  private int get_higher_power_of_two(int number) {
    assert (number < 100000);
    int result = 1;
    while (result < number) {
      result = result << 1;
    }
    return result;
  }


  // Adds byes to the given player list to bring size up to next power of 2
  private void add_byes(ArrayList p_list) {
    int size = p_list.size();
    int power2 = get_higher_power_of_two(size);
    for (int i = size; i < power2; i++) {
      p_list.add(bye);
    }
  }

  // phase of the tournament
  enum PHASE {
    PRELIM, SEMI, FINAL} ;

    // Runs a double elim tournament
    private void tourny_double_elim(int semi_wins, int final_wins, int winner_advantage) {

      // Fill out players list
      add_byes(player_list);

      // Dump out current tournament state
 //     LogFile.message(player_list.toString());
 //     LogFile.message(match_list.toString());

      // Create a list of winners and losers
      ArrayList winner_list = new ArrayList();
      ArrayList loser_list = new ArrayList();
      ArrayList eliminated_list = new ArrayList();
      winner_list.addAll(player_list);

      PHASE cur_stage = PHASE.PRELIM;
      Player third_place = bye;

      // Process the already played games
      for (int i = 0; i < match_list.size(); i++) {

        // Determine tourny phase
        if (winner_list.size() == 1 && loser_list.size() == 2) {
          cur_stage = PHASE.SEMI;
        } else if (winner_list.size() == 1 && loser_list.size() == 1) {
          cur_stage = PHASE.FINAL;
        } else {
          cur_stage = PHASE.PRELIM;
        }

        MatchResult match = (MatchResult) match_list.get(i);
        Player loser = match.loser;
        Player winner = match.winner;

        switch (cur_stage) {

        case PRELIM:
          // Send 1st time losers to the losers bracket
          if (winner_list.contains(loser)) {
            winner_list.remove(loser);
            loser_list.add(loser);
          }

          // Send 2nd time losers home
          else if (loser_list.contains(loser)) {
            loser_list.remove(loser);
            eliminated_list.add(loser);
          }
          break;

        case SEMI:
          winner.wins++;
          if ( winner.wins == semi_wins ) {
            loser_list.remove(loser);
            eliminated_list.add(loser);
            winner.wins = 0;  // reset win count for finals
            ((Player)winner_list.get(0)).wins=winner_advantage;  // Winners side advantage in final
            third_place = loser;  // Store third place finisher
          }
          break;

        case FINAL:
          winner.wins++;
          if ( winner.wins == final_wins ) {
            System.out.println("result " + third_place.name + " 3");
            System.out.println("result " + loser.name + " 2");
            System.out.println("result " + winner.name + " 1");
            eliminated_list.add(loser);
          }
          break;
        }
      }

//      LogFile.message("Phase: "+cur_stage);
//      LogFile.message("W: " + winner_list.toString());
//      LogFile.message("L: " + loser_list.toString());
//      LogFile.message("E: " + eliminated_list.toString());

      // Check for the winner
      if (eliminated_list.size() == player_list.size() - 1) {
        // do nothing!!!
        return;
      }

      // Check for final two players left
      if (eliminated_list.size() == player_list.size() - 2) {
        player_list.removeAll(eliminated_list);
        Player one = (Player) player_list.get(0);
        Player two = (Player) player_list.get(1);
        System.out.println("pick " + one.name + " " + two.name);
        return;
      }

      // Generate pairings
      if (winner_list.size() > loser_list.size()) {
        generate_pairings(winner_list);
      } else {
        generate_pairings(loser_list);
      }

    }

    // Generates pairings for the given list
    private void generate_pairings(ArrayList p_list) {

      // Add byes to bring list up to a good length
      add_byes(p_list);

      Collections.sort(p_list);
//    LogFile.message("Generating Pairings\n");
//    LogFile.message(p_list.toString());

      for (int i = 0; i < p_list.size() / 2; i++) {
        int p1 = i;
        int p2 = p_list.size() - 1 - i; // Fold Method

        Player one = (Player) p_list.get(p1);
        Player two = (Player) p_list.get(p2);

        System.out.println("pick " + one.name + " " + two.name);
      }
    }


    // Runs a single elimination tournament
    private void tourny_single_elim() {

      // Dump out current tournament state
      //LogFile.message(player_list.toString());
      //LogFile.message(match_list.toString());

      // Remove the losers
      for (int i = 0; i < match_list.size(); i++) {
        MatchResult match = (MatchResult) match_list.get(i);
        player_list.remove(match.loser);
      }

      // Check for the winner
      if (player_list.size() == 1) {
        Player winner = (Player) player_list.get(0);
        System.out.println("result " + winner.name + " 1");
        return;
      }

      // Generate pairings
      generate_pairings(player_list);

    }


    public static void main(String args[]) {

      boolean use_test_version = false;

      try {

        if (use_test_version) {
//        LogFile.setMessageDisplay(true);
          TournamentSimulation sim = new TournamentSimulation("ts.4032");
          sim.tourny_double_elim(1,2,1);
          return;
        }

        // Check if we got a tournament state file
        if (args.length < 1) {
          System.out.println("Error: No Tournament State File!");
          return;
        }

//        LogFile.setMessageDisplay(false);
        TournamentSimulation sim = new TournamentSimulation(args[0]);
        int semi_wins = Integer.parseInt(args[1]);
        int final_wins = Integer.parseInt(args[2]);
        int winner_advantage = Integer.parseInt(args[3]);
        sim.tourny_double_elim(semi_wins,final_wins,winner_advantage);

      } catch (Exception e) {
        System.out.println("Error: " + e);
      }
    }

  }
