#ifndef COLORS_HPP
#define COLORS_HPP

#include <map>

class ColorHistory {
  std::map<std::string,int> goldMinusSilver;
  std::map<std::string,int> streak;
  std::map<std::pair<std::string,std::string>,int> goldMinusSilverVersusSecondPlayer;
  std::map<std::pair<std::string,std::string>,int> lastColorFirstPlayer;
public:
  void addGame(const std::string& goldPlayer,const std::string& silverPlayer)
  {
    ++goldMinusSilver[goldPlayer];
    --goldMinusSilver[silverPlayer];
    if (goldPlayer<silverPlayer) {
      const std::pair<std::string,std::string> sortedPair=make_pair(goldPlayer,silverPlayer);
      ++goldMinusSilverVersusSecondPlayer[sortedPair];
      lastColorFirstPlayer[sortedPair]=1;
    }
    else {
      const std::pair<std::string,std::string> sortedPair=make_pair(silverPlayer,goldPlayer);
      --goldMinusSilverVersusSecondPlayer[sortedPair];
      lastColorFirstPlayer[sortedPair]=-1;
    }
    if (streak[goldPlayer]>0)
      ++streak[goldPlayer];
    else
      streak[goldPlayer]=1;
    if (streak[silverPlayer]<0)
      --streak[silverPlayer];
    else
      streak[silverPlayer]=-1;
  }

  int colorFirstPlayer(const std::string& firstPlayer,const std::string& secondPlayer)
  {
    if (goldMinusSilver[firstPlayer]!=goldMinusSilver[secondPlayer])
      return goldMinusSilver[secondPlayer]-goldMinusSilver[firstPlayer];

    const std::pair<std::string,std::string> sortedPair=(firstPlayer<secondPlayer) ?
      make_pair(firstPlayer,secondPlayer) :
      make_pair(secondPlayer,firstPlayer);

    if (goldMinusSilverVersusSecondPlayer[sortedPair]!=0)
      return (firstPlayer==sortedPair.first) ?
        -goldMinusSilverVersusSecondPlayer[sortedPair] :
         goldMinusSilverVersusSecondPlayer[sortedPair];

    if (streak[firstPlayer]!=streak[secondPlayer])
      return streak[secondPlayer]-streak[firstPlayer];

    return (firstPlayer==sortedPair.first) ?
      -lastColorFirstPlayer[sortedPair] :
       lastColorFirstPlayer[sortedPair];
  }
};

#endif

